/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Prepaid calling card application
 * 
 * Copyright (C) 2003, Green Tone N.V. and Igor de Rooij
 *
 * $Id: app_prepaid.c,v 1.17 2004/09/28 14:06:00 tuyano Exp $
 *
 * TODO:
 * - (simpler queries/db struct for release ( unnormalized card,rate etc))
 * - better error checks
 * - TEST TEST TEST!
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 *
 * - bug where authentication results are also in authorization results, added postgreslock to fix
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/app.h>
#include <asterisk/say.h>

#include <asterisk/config.h>
#include <pgsql/libpq-fe.h>

#include <asterisk/translate.h>
#include <asterisk/musiconhold.h>
#include <asterisk/callerid.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <asterisk/options.h>

#define DATE_FORMAT "%Y-%m-%d %T"

static char *tdesc = "Prepaid Application";

static char *app = "Prepaid";

static char *config = "prepaid.conf";

static char *synopsis = "Prepaid Authentication, Authorization and Accounting";

static char *descrip =
"  PrepaidIVR: Requires a user to enter a"
" 10 digit cardnumber. also authorizes the called\n"
" destination and handles accounting\n";

static char *pghostname = NULL, *pgdbname = NULL, *pgdbuser = NULL, *pgpassword = NULL, *pgdbsock = NULL, *pgdbport = NULL;
static char *psqlauthentication = NULL, *psqlauthorization = NULL, *psqlacctstart = NULL, *psqlacctstop = NULL, *psqlsimulcount = NULL;
static int hostname_alloc = 0, dbname_alloc = 0, dbuser_alloc = 0, password_alloc = 0, dbsock_alloc = 0, dbport_alloc = 0;
static int sqlauthentication_alloc = 0, sqlauthorization_alloc = 0, sqlacctstart_alloc = 0, sqlacctstop_alloc = 0, sqlsimulcount_alloc = 0;
static int connected = 0;

PGconn *conn;
ast_mutex_t postgreslock;

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

struct calldata {
    char cardnumber[256];
    char destination[256];
    char *credit;
    char *tariff;
    char *active;
    char *rate;
    char *tech;
    char *prefix;
    char *hostname;
    int timeout;
    char newdestination[256];
};

// Checks if we are connected to the db - if not then it tries to reconnect
static int check_connected(void)
{
        int i;
        PGresult *result;

        if (connected) {
                // seems that we are connected - but let us check that
                for (i=0;i<3;i++) {
                        // PQstatus is not working as expected - so we are trying a query before PQstatus - hope that helps
                        result = PQexec(conn, "SELECT 1");
                        PQclear(result);
                        if (PQstatus(conn)==CONNECTION_BAD) {
                                // seems that we are having troubles with the connection - trying to reset the connection
                                // and set connected to false
                                connected = 0;
                                ast_log(LOG_ERROR, "app_prepaid: connection to the database server %s failed - trying to reconnect (3 times)\n", pghostname
);
                                PQreset(conn);
                                // now retry it (3 times)
                                continue;
                        } else {
                                // Everything is fine - so go on
                                connected = 1;
                                break;
                        }
                }
        } else {
                // We are not connected - so make the initial connection
                conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
                if (PQstatus(conn) != CONNECTION_BAD) {
                        ast_log(LOG_DEBUG,"Successfully connected to Prepaid database.\n");
                        connected = 1;
                } else {
                        ast_log(LOG_ERROR, "app_prepaid: cannot connect to database server %s.  Calls will not be logged\n", pghostname);
                        connected = 0;
                }
        }
        return connected;
}


static int prepaid_ivr_authenticate(struct ast_channel *chan, struct calldata *cd)
{
    int res=0;
    int retries;
    // char cardnumber[256] = "";
    char *simulcount = NULL;
    char *prompt = NULL, *username = NULL, *userpass = NULL;
    char sqlcmd[2048] = "";
    PGresult *result;

    prompt = "prepaid-enter-card-num";
    for (retries = 0; retries < 3; retries++) {
	res = ast_app_getdata(chan, prompt, cd->cardnumber, sizeof(cd->cardnumber) - 2, 0);
        if (res < 0) {
	    res = -1;
    	    break;
	}
        res = 0;

	if ( strlen(cd->cardnumber) == 0) {
	    prompt = "prepaid-no-card-entered";
	    continue;
	}
        
	if ( strlen(cd->cardnumber) != 10) {
	    prompt = "prepaid-invalid-digits";
	    continue;
	}

        username = (char *)strndup(cd->cardnumber,5);
        userpass = &cd->cardnumber[5];
        sprintf(sqlcmd,psqlauthentication,username, userpass);

        ast_log(LOG_DEBUG,"app_prepaid: SQL Authenticate command as follows:  %s\n",sqlcmd);
	ast_mutex_lock(&postgreslock);
        result = PQexec(conn, sqlcmd);
        if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
            ast_log(LOG_DEBUG,"app_prepaid: FATAL ERROR: %s\n", PQresultErrorMessage(result));
            prompt="prepaid-no-aaa"; // tech error sql failed
	    ast_mutex_unlock(&postgreslock);
	    res = -2;
            break;
        }
        if(PQntuples(result)<=0) {
            prompt="prepaid-auth-fail";
	    ast_mutex_unlock(&postgreslock);
            continue;
        }
	
	// use srncopy here...
        cd->credit = PQgetvalue(result,0,0);
        cd->tariff = PQgetvalue(result,0,1);
        cd->active = PQgetvalue(result,0,2);
        ast_log(LOG_DEBUG,"app_prepaid: SQL Authenticate results:  credit='%s', tariff='%s', activated='%s'\n",cd->credit,cd->tariff,cd->active);

	// check credit > 0 / you have zero balance
        if( atoi(cd->credit) <= 0 ) {
	    prompt = "prepaid-zero-balance";
	    ast_mutex_unlock(&postgreslock);
	    res = -2;
            break;
	}
        
	// check activated=t / card not active, contact customer support
	if(strcmp(cd->active, "t") != 0 ) {
	    // not expired but inactive.. probably not yet sold.. find better prompt
	    prompt = "prepaid-card-expired";
	    ast_mutex_unlock(&postgreslock);
	    res = -2;
            break;
	}
	
	// check double logins..
	sprintf(sqlcmd,"select count(*) from call where username='%s' and stoptime is null", username);
        result = PQexec(conn, sqlcmd);
        if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
            ast_log(LOG_DEBUG,"app_prepaid: FATAL ERROR: %s\n", PQresultErrorMessage(result));
            prompt="prepaid-no-aaa"; // tech error sql failed
	    ast_mutex_unlock(&postgreslock);
            res = -2;
            break;
        }
	simulcount = PQgetvalue(result,0,0);
        if( strcmp(simulcount,"0") != 0 ) {
            ast_log(LOG_DEBUG,"app_prepaid: WARNING: Card number %s is already in use!\n", username);
            prompt="prepaid-card-in-use";
	    ast_mutex_unlock(&postgreslock);
	    res = -2;
            break;
        }
	
	ast_mutex_unlock(&postgreslock);
        break;
    }
    
    if ((retries < 3) && !res) {
	int dollars, cents;
	div_t result;
        ast_cdr_setaccount(chan, username);

        // say 'you have x dollars and x cents'
	result = div (atoi(cd->credit),100);
	
	dollars = result.quot;
	cents = result.rem;
	
        res = ast_streamfile(chan, "prepaid-you-have", chan->language);
        if (!res)
            res = ast_waitstream(chan, "");
	
	if (result.quot > 0) {
    	    res = ast_say_number(chan, result.quot, "", chan->language, (char *) NULL);
    	    if (!res)
        	res = ast_waitstream(chan, "");
	}
	
	if ( result.quot == 1 ) {
    	    res = ast_streamfile(chan, "digits/dollar", chan->language);    
    	    if (!res)
        	res = ast_waitstream(chan, "");
		
	} else if (result.quot > 1) {
    	    res = ast_streamfile(chan, "digits/dollars", chan->language);
    	    if (!res)
        	res = ast_waitstream(chan, "");
		
	}

	if (result.quot > 0 && result.rem > 0) {
    	    res = ast_streamfile(chan, "vm-and", chan->language);
    	    if (!res)
        	res = ast_waitstream(chan, "");
	}
	
	if ( result.rem > 0) {
    	    res = ast_say_number(chan, result.rem, "", chan->language, (char *) NULL);
    	    if (!res)
        	res = ast_waitstream(chan, "");
    	    res = ast_streamfile(chan, "prepaid-cents", chan->language);
    	    if (!res)
        	res = ast_waitstream(chan, "");
	    res = 0;
	}
	
    } else if (res == -2 ) {
        res = ast_streamfile(chan, prompt, chan->language);
        if (!res)
            res = ast_waitstream(chan, "");
	res = -1;
    } else {
        res = ast_streamfile(chan, "prepaid-final", chan->language);
        if (!res)
            res = ast_waitstream(chan, "");
	res = -1;
    }

    return res;

}

static int prepaid_ivr_authorize(struct ast_channel *chan, struct calldata *cd)
{
    int res=0;
    int retries;
    // char destination[256] = "";
    char *prompt = NULL;
    char sqlcmd[2048] = "";
    PGresult *result;


    // ask destination
    prompt = "prepaid-enter-dest";
    for (retries = 0; retries < 3; retries++) {
        res = ast_app_getdata(chan, prompt, cd->destination, sizeof(cd->destination) - 2, 0);
        if (res < 0)
            break;
        res = 0;

        // lookup rate
        sprintf(sqlcmd,psqlauthorization,cd->tariff,cd->destination);
        ast_log(LOG_DEBUG,"app_prepaid: SQL Authorize command as follows:  %s\n",sqlcmd);
	ast_mutex_lock(&postgreslock);
        result = PQexec(conn, sqlcmd);
        if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
            ast_log(LOG_DEBUG,"app_prepaid: FATAL ERROR: %s\n", PQresultErrorMessage(result));
            prompt="prepaid-no-aaa"; // tech error sql failed
	    ast_mutex_unlock(&postgreslock);
	    //reenter dest
            continue;
        }
        if(PQntuples(result)<=0) {
            ast_log(LOG_DEBUG,"app_prepaid: WARNING: No destination route for \"%s\"\n", cd->destination);
            prompt="prepaid-dest-unreachable";
	    ast_mutex_unlock(&postgreslock);	
            continue;
        }
        cd->rate = PQgetvalue(result,0,0);
        if ( !cd->rate || !strlen(cd->rate) ) {
            ast_log(LOG_DEBUG,"app_prepaid: WARNING: Destination \"%s\" is blocked.\n", cd->destination);
            prompt="prepaid-dest-blocked";
	    ast_mutex_unlock(&postgreslock);
            continue;
        }
	if( atoi(cd->rate) <= 0 ) {
            ast_log(LOG_DEBUG,"app_prepaid: WARNING: Negative rate for destination \"%s\"\n", cd->destination);
	    prompt="prepaid-dest-unreachable";
	    ast_mutex_unlock(&postgreslock);
	    continue;
        }
        cd->tech = PQgetvalue(result,0,1);
        cd->prefix = PQgetvalue(result,0,2);
        cd->hostname = PQgetvalue(result,0,3);
	strncpy(cd->newdestination, PQgetvalue(result,0,4), sizeof(cd->newdestination));

        // checks !! rate != null a.o.
	

        ast_log(LOG_DEBUG,"app_prepaid: SQL Authorize results:  rate='%s', tech='%s', prefix='%s', hostname='%s'\n",cd->rate, cd->tech, cd->prefix, cd->hostname);
	ast_mutex_unlock(&postgreslock);
        break;
    }
    if ((retries < 3) && !res) {
	div_t result;
	
	ast_cdr_setapp(chan->cdr, app, cd->newdestination);

        // calculate timeout
        cd->timeout = atoi(cd->credit) * 60 / atoi(cd->rate);
        ast_log(LOG_DEBUG,"app_prepaid: timeout='%i'\n",cd->timeout);

        // set destination and timeout
        // say 'you have x minutes and x seconds'
	
	result = div (cd->timeout,60);

        ast_log(LOG_DEBUG,"app_prepaid: minutes='%i', seconds='%i'\n",result.quot, result.rem);
	
        res = ast_streamfile(chan, "prepaid-you-have", chan->language);
        if (!res)
            res = ast_waitstream(chan, "");

	if (result.quot > 0) {
    	    res = ast_say_number(chan, result.quot, "", chan->language, (char *) NULL);
    	    if (!res)
        	res = ast_waitstream(chan, "");
	}
	
	if ( result.quot == 1 ) {
    	    res = ast_streamfile(chan, "prepaid-minute", chan->language);    
    	    if (!res)
        	res = ast_waitstream(chan, "");
		
	} else if (result.quot > 1) {
    	    res = ast_streamfile(chan, "prepaid-minutes", chan->language);
    	    if (!res)
        	res = ast_waitstream(chan, "");
		
	}

	if (result.quot > 0 && result.rem > 0) {
    	    res = ast_streamfile(chan, "vm-and", chan->language);
    	    if (!res)
        	res = ast_waitstream(chan, "");
	}
	
	if (result.rem > 0) {
    	    if (!res)
        	res = ast_waitstream(chan, "");
    	    res = ast_say_number(chan, result.rem, "", chan->language, (char *) NULL);
    	    if (!res)
        	res = ast_waitstream(chan, "");
    	    res = ast_streamfile(chan, "prepaid-seconds", chan->language);
    	    if (!res)
        	res = ast_waitstream(chan, "");
	    res = 0;
	}

    } else {
	// auth failed
        res = ast_streamfile(chan, "prepaid-final", chan->language);
        if (!res)
            res = ast_waitstream(chan, "");
	res = -1;
    }
    
    return res;

}

static int prepaid_pbx_dial(struct ast_channel *chan, void *data)
{
    struct ast_app *app;
    int ret;

    app = pbx_findapp("Dial");
    if (app) {
        ret = pbx_exec(chan, app, data, 1);
    } else {
        ast_log(LOG_WARNING, "Could not find application (Dial)\n");
        ret = -2;
    }
    return ret;
}

static int prepaid_acct_start(struct ast_cdr *cdr, struct calldata *cd)
{
    struct tm tm;
    struct timeval tv;
    char sqlcmd[2048], timestr[128];
    time_t t;
    PGresult *result;

    memset(sqlcmd,0,2048);

    gettimeofday(&tv,NULL);
    t = tv.tv_sec;
    localtime_r(&t,&tm);
    strftime(timestr,128,DATE_FORMAT,&tm);

    //TODO: add some sql reconnect stuff here ( also on auth/autz)

    if (connected) {
        char *clid=NULL, *dcontext=NULL, *channel=NULL, *dstchannel=NULL, *lastapp=NULL, *lastdata=NULL;
        char *uniqueid=NULL;
	char hostname[256];
	
        if (gethostname(hostname, sizeof(hostname)))
	    strncpy(hostname, "<Unknown>", sizeof(hostname)-1);
			
        /* Maximum space needed would be if all characters needed to be escaped, plus a trailing NULL */
        if ((clid = alloca(strlen(cdr->clid) * 2 + 1)) != NULL)
            PQescapeString(clid, cdr->clid, strlen(cdr->clid));
        if ((dcontext = alloca(strlen(cdr->dcontext) * 2 + 1)) != NULL)
            PQescapeString(dcontext, cdr->dcontext, strlen(cdr->dcontext));
        if ((channel = alloca(strlen(cdr->channel) * 2 + 1)) != NULL)
            PQescapeString(channel, cdr->channel, strlen(cdr->channel));
        if ((dstchannel = alloca(strlen(cdr->dstchannel) * 2 + 1)) != NULL)
            PQescapeString(dstchannel, cdr->dstchannel, strlen(cdr->dstchannel));
        if ((lastapp = alloca(strlen(cdr->lastapp) * 2 + 1)) != NULL)
            PQescapeString(lastapp, cdr->lastapp, strlen(cdr->lastapp));
        if ((lastdata = alloca(strlen(cdr->lastdata) * 2 + 1)) != NULL)
            PQescapeString(lastdata, cdr->lastdata, strlen(cdr->lastdata));
        if ((uniqueid = alloca(strlen(cdr->uniqueid) * 2 + 1)) != NULL)
            PQescapeString(uniqueid, cdr->uniqueid, strlen(cdr->uniqueid));

        /* Check for all alloca failures above at once */
        if ((!clid) || (!dcontext) || (!channel) || (!dstchannel) || (!lastapp) || (!lastdata) || (!uniqueid)) {
            ast_log(LOG_ERROR, "app_prepaid:  Out of memory error (insert fails)\n");
            // ast_mutex_unlock(&prepaid_cdr_lock);
            return -1;
        }
    
	ast_log(LOG_DEBUG,"app_prepaid: Inserting Prepaid Call record.\n");
	//( uniqueid, sessionid, username, nasipaddress, starttime)
	sprintf(sqlcmd,psqlacctstart,uniqueid, channel, cdr->accountcode, hostname, timestr);

	ast_log(LOG_DEBUG,"app_prepaid: SQL command as follows:  %s\n",sqlcmd);
	ast_mutex_lock(&postgreslock);
	result = PQexec(conn, sqlcmd);
	if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
    	    ast_log(LOG_ERROR,"app_prepaid: Failed to insert accounting record.\n");
            ast_log(LOG_DEBUG,"app_prepaid: ERROR: %s\n", PQresultErrorMessage(result));
	    ast_mutex_unlock(&postgreslock);
    	    return -1;
	}
	ast_mutex_unlock(&postgreslock);
    }

    return 0;
}

static int prepaid_acct_stop(struct ast_cdr *cdr, struct calldata *cd)
{
    struct tm tm;
    struct timeval tv;
    char sqlcmd[2048], timestr[128];
    time_t t;
    PGresult *result;

    memset(sqlcmd,0,2048);

    gettimeofday(&tv,NULL);
    t = tv.tv_sec;
    localtime_r(&t,&tm);
    strftime(timestr,128,DATE_FORMAT,&tm);

    ast_cdr_end(cdr);

    if (!cdr->end.tv_sec && !cdr->end.tv_usec)
        ast_log(LOG_WARNING, "app_prepaid: CDR on channel '%s' lacks end\n", cdr->channel);
    if (!cdr->start.tv_sec && !cdr->start.tv_usec)
        ast_log(LOG_WARNING, "app_prepaid: CDR on channel '%s' lacks start\n", cdr->channel);
    cdr->duration = cdr->end.tv_sec - cdr->start.tv_sec + (cdr->end.tv_usec - cdr->start.tv_usec) / 1000000;
    if (cdr->answer.tv_sec || cdr->answer.tv_usec) {
        cdr->billsec = cdr->end.tv_sec - cdr->answer.tv_sec + (cdr->end.tv_usec - cdr->answer.tv_usec) / 1000000;
    } else
        cdr->billsec = 0;

    //TODO: add some sql reconnect stuff here ( also on auth/autz)

    if (connected) {
        char *clid=NULL, *dcontext=NULL, *channel=NULL, *dstchannel=NULL, *lastapp=NULL, *lastdata=NULL;
        char *uniqueid=NULL;

        /* Maximum space needed would be if all characters needed to be escaped, plus a trailing NULL */
        if ((clid = alloca(strlen(cdr->clid) * 2 + 1)) != NULL)
            PQescapeString(clid, cdr->clid, strlen(cdr->clid));
        if ((dcontext = alloca(strlen(cdr->dcontext) * 2 + 1)) != NULL)
            PQescapeString(dcontext, cdr->dcontext, strlen(cdr->dcontext));
        if ((channel = alloca(strlen(cdr->channel) * 2 + 1)) != NULL)
            PQescapeString(channel, cdr->channel, strlen(cdr->channel));
        if ((dstchannel = alloca(strlen(cdr->dstchannel) * 2 + 1)) != NULL)
            PQescapeString(dstchannel, cdr->dstchannel, strlen(cdr->dstchannel));
        if ((lastapp = alloca(strlen(cdr->lastapp) * 2 + 1)) != NULL)
            PQescapeString(lastapp, cdr->lastapp, strlen(cdr->lastapp));
        if ((lastdata = alloca(strlen(cdr->lastdata) * 2 + 1)) != NULL)
            PQescapeString(lastdata, cdr->lastdata, strlen(cdr->lastdata));
        if ((uniqueid = alloca(strlen(cdr->uniqueid) * 2 + 1)) != NULL)
            PQescapeString(uniqueid, cdr->uniqueid, strlen(cdr->uniqueid));

        /* Check for all alloca failures above at once */
        if ((!clid) || (!dcontext) || (!channel) || (!dstchannel) || (!lastapp) || (!lastdata) || (!uniqueid)) {
            ast_log(LOG_ERROR, "app_prepaid:  Out of memory error (insert fails)\n");
            // ast_mutex_unlock(&prepaid_cdr_lock);
            return -1;
        }
    
	ast_log(LOG_DEBUG,"app_prepaid: Updating Prepaid Call record.\n");
	//stoptime = '%s', sessiontime = '%i', sessionbill = '%i', terminatecause = '%s',uniqueid='%s'
	
	sprintf(sqlcmd,psqlacctstop,uniqueid, channel, cdr->accountcode, cd->destination, atoi(cd->rate), cdr->billsec, ast_cdr_disp2str(cdr->disposition), timestr);
	
	ast_log(LOG_DEBUG,"app_prepaid: SQL command as follows:  %s\n",sqlcmd);
	ast_mutex_lock(&postgreslock);
	result = PQexec(conn, sqlcmd);
	if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
    	    ast_log(LOG_ERROR,"app_prepaid: Failed to update accounting record.\n");
            ast_log(LOG_DEBUG,"app_prepaid: ERROR: %s\n", PQresultErrorMessage(result));
	    ast_mutex_unlock(&postgreslock) ;
    	    return -1;
	}
	ast_mutex_unlock(&postgreslock);
    }

    return 0;
}

static int prepaid_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	struct localuser *u;
	struct calldata cd;
	char dialstr[256] = "";

	LOCAL_USER_ADD(u);
	if (chan->_state != AST_STATE_UP) {
		res = ast_answer(chan);
		if (res) {
			LOCAL_USER_REMOVE(u);
			return -1;
		}
	}
	

	res = prepaid_ivr_authenticate(chan, &cd);
	if (!res) {
	
	    prepaid_acct_start(chan->cdr, &cd);

	    res = prepaid_ivr_authorize(chan, &cd);
	    if (!res) {
                ast_cdr_reset(chan->cdr,0);
                ast_channel_setwhentohangup(chan,cd.timeout);

                sprintf(dialstr,"%s/%s%s@%s", cd.tech, cd.prefix, chan->cdr->lastdata, cd.hostname);
                ast_log(LOG_DEBUG,"app_prepaid: Dialing '%s' with timeout of '%d'.\n",dialstr,cd.timeout);
                res = prepaid_pbx_dial(chan, dialstr);
		
	    }
	    
	    prepaid_acct_stop(chan->cdr, &cd);
	    
	}
	
	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
        STANDARD_HANGUP_LOCALUSERS;
        PQfinish(conn);
        connected = 0;
        if (pghostname && hostname_alloc) {
                free(pghostname);
                pghostname = NULL;
                hostname_alloc = 0;
        }
        if (pgdbname && dbname_alloc) {
                free(pgdbname);
                pgdbname = NULL;
                dbname_alloc = 0;
        }
        if (pgdbuser && dbuser_alloc) {
                free(pgdbuser);
                pgdbuser = NULL;
                dbuser_alloc = 0;
        }
        if (pgdbsock && dbsock_alloc) {
                free(pgdbsock);
                pgdbsock = NULL;
                dbsock_alloc = 0;
        }
        if (pgpassword && password_alloc) {
                free(pgpassword);
                pgpassword = NULL;
                password_alloc = 0;
        }
        if (pgdbport && dbport_alloc) {
                free(pgdbport);
                pgdbport = NULL;
                dbport_alloc = 0;
        }
        if (psqlauthentication && sqlauthentication_alloc) {
                free(psqlauthentication);
                psqlauthentication = NULL;
                sqlauthentication_alloc = 0;
        }
        if (psqlauthorization && sqlauthorization_alloc) {
                free(psqlauthorization);
                psqlauthorization = NULL;
                sqlauthorization_alloc = 0;
        }
        if (psqlacctstart && sqlacctstart_alloc) {
                free(psqlacctstart);
                psqlacctstart = NULL;
                sqlacctstart_alloc = 0;
        }
        if (psqlacctstop && sqlacctstop_alloc) {
                free(psqlacctstop);
                psqlacctstop = NULL;
                sqlacctstop_alloc = 0;
        }
        if (psqlsimulcount && sqlsimulcount_alloc) {
                free(psqlsimulcount);
                psqlsimulcount = NULL;
                sqlsimulcount_alloc = 0;
        }

	ast_unregister_application(app);

	return 0;
}

int load_module(void)
{
        int res;
        struct ast_config *cfg;
        struct ast_variable *var;
        char *tmp;

        cfg = ast_config_load(config);
        if (!cfg) {
                ast_log(LOG_WARNING, "Unable to load config for Prepaid Application: %s\n", config);
                return 0;
        }

        var = ast_variable_browse(cfg, "global");
        if (!var) {
                /* nothing configured */
                return 0;
        }

        tmp = ast_variable_retrieve(cfg,"global","hostname");
        if (tmp) {
                pghostname = malloc(strlen(tmp) + 1);
                if (pghostname != NULL) {
                        hostname_alloc = 1;
			strncpy(pghostname, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database server hostname not specified.  Assuming localhost\n");
                pghostname = "localhost";
        }

        tmp = ast_variable_retrieve(cfg,"global","dbname");
        if (tmp) {
                pgdbname = malloc(strlen(tmp) + 1);
                if (pgdbname != NULL) {
                        dbname_alloc = 1;
			strncpy(pgdbname, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database not specified.  Assuming asterisk\n");
                pgdbname = "asterisk";
        }

        tmp = ast_variable_retrieve(cfg,"global","user");
        if (tmp) {
                pgdbuser = malloc(strlen(tmp) + 1);
                if (pgdbuser != NULL) {
                        dbuser_alloc = 1;
			strncpy(pgdbuser, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database user not specified.  Assuming postgres\n");
                pgdbuser = "postgres";
        }

        tmp = ast_variable_retrieve(cfg,"global","password");
        if (tmp) {
                pgpassword = malloc(strlen(tmp) + 1);
                if (pgpassword != NULL) {
                        password_alloc = 1;
			strncpy(pgpassword, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database password not specified.  Assuming blank\n");
                pgpassword = "";
        }

        tmp = ast_variable_retrieve(cfg,"global","port");
        if (tmp) {
                pgdbport = malloc(strlen(tmp) + 1);
                if (pgdbport != NULL) {
                        dbport_alloc = 1;
			strncpy(pgdbport, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database port not specified.  Using 5432.\n");
                pgdbport = "5432";
        }


        tmp = ast_variable_retrieve(cfg,"sql","authentication");
        if (tmp) {
                psqlauthentication = malloc(strlen(tmp) + 1);
                if (psqlauthentication != NULL) {
                        sqlauthentication_alloc = 1;
			strncpy(psqlauthentication, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid authentication query not specified.  Using default.\n");
                psqlauthentication = "SELECT credit, tariff, activated FROM card WHERE username='%s' and userpass='%s'";
        }

        tmp = ast_variable_retrieve(cfg,"sql","authorization");
        if (tmp) {
                psqlauthorization = malloc(strlen(tmp) + 1);
                if (psqlauthorization != NULL) {
                        sqlauthorization_alloc = 1;
			strncpy(psqlauthorization, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid authorization query not specified.  Using default.\n");
                psqlauthorization = "SELECT rate from countryprefix LEFT JOIN tariffrate USING (countrycode,subcode) WHERE tariffcode='%s' AND prefix=SUBSTRING('%s', 1, length(prefix)) ORDER BY LENGTH(prefix) desc";
        }


        tmp = ast_variable_retrieve(cfg,"sql","acct_start");
        if (tmp) {
                psqlacctstart = malloc(strlen(tmp) + 1);
                if (psqlacctstart != NULL) {
                        sqlacctstart_alloc = 1;
			strncpy(psqlacctstart, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid acct_start query not specified.  Using default.\n");
                psqlacctstart = "INSERT INTO cdr (calldate,clid,src,dst,dcontext,channel,dstchannel,lastapp,lastdata,duration,billsec,disposition,amaflags,accountcode,uniqueid) VALUES ('%s','%s','%s','%s','%s', '%s','%s','%s','%s',%i,%i,'%s',%i,'%s','%s')";
        }

        tmp = ast_variable_retrieve(cfg,"sql","acct_stop");
        if (tmp) {
                psqlacctstop = malloc(strlen(tmp) + 1);
                if (psqlacctstop != NULL) {
                        sqlacctstop_alloc = 1;
			strncpy(psqlacctstop, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid acct_stop query not specified.  Using default.\n");
                psqlacctstop = "INSERT INTO cdr (calldate,clid,src,dst,dcontext,channel,dstchannel,lastapp,lastdata,duration,billsec,disposition,amaflags,accountcode,uniqueid) VALUES ('%s','%s','%s','%s','%s', '%s','%s','%s','%s',%i,%i,'%s',%i,'%s','%s')";
        }

        tmp = ast_variable_retrieve(cfg,"sql","simulcount");
        if (tmp) {
                psqlsimulcount = malloc(strlen(tmp) + 1);
                if (psqlsimulcount != NULL) {
                        sqlsimulcount_alloc = 1;
			strncpy(psqlsimulcount, tmp, strlen(tmp)+1);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid simulcount query not specified.  Using default.\n");
                psqlsimulcount = "select count(*) from call where username='%s' and stoptime is null;";
        }

        ast_config_destroy(cfg);

        ast_log(LOG_DEBUG,"app_prepaid: got hostname of %s\n",pghostname);
        ast_log(LOG_DEBUG,"app_prepaid: got port of %s\n",pgdbport);
        if (pgdbsock)
                ast_log(LOG_DEBUG,"app_prepaid: got sock file of %s\n",pgdbsock);
        ast_log(LOG_DEBUG,"app_prepaid: got user of %s\n",pgdbuser);
        ast_log(LOG_DEBUG,"app_prepaid: got dbname of %s\n",pgdbname);
        ast_log(LOG_DEBUG,"app_prepaid: got password of %s\n",pgpassword);

        ast_log(LOG_DEBUG,"app_prepaid: got sqlauthentication of '%s'\n",psqlauthentication);
        ast_log(LOG_DEBUG,"app_prepaid: got sqlauthorization of '%s'\n",psqlauthorization);
        ast_log(LOG_DEBUG,"app_prepaid: got sqlacct_start of '%s'\n",psqlacctstart);
        ast_log(LOG_DEBUG,"app_prepaid: got sqlacct_stop of '%s'\n",psqlacctstop);
        ast_log(LOG_DEBUG,"app_prepaid: got sqlsimulcount of '%s'\n",psqlsimulcount);

        conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
        if (PQstatus(conn) != CONNECTION_BAD) {
                ast_log(LOG_DEBUG,"Successfully connected to Prepaid database.\n");
                connected = 1;
        } else {
                ast_log(LOG_ERROR, "app_prepaid: cannot connect to database server %s.  Calls will not be logged\n", pghostname);
                connected = 0;
        }
	ast_mutex_init(&postgreslock);

        res = ast_register_application(app, prepaid_exec, synopsis, descrip);
        if (res) {
                ast_log(LOG_ERROR, "Unable to register Prepaid Application\n");
        }

        return res;
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}

