/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Prepaid Rate application
 * 
 * Copyright (C) 2004, Peter Nixon
 *
 * Peter Nixon <codemonkey@peternixon.net>
 *
 * $Id: app_prepaid_rate.c,v 1.16 2004/09/28 14:06:00 tuyano Exp $
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#define sqlcommandsize 2048

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <asterisk/utils.h>

#include <asterisk/app.h>
#include <asterisk/say.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>
#include  <math.h>

#include <asterisk/config.h>
#include <pgsql/libpq-fe.h>

#include <asterisk/translate.h>
#include <asterisk/musiconhold.h>
#include <asterisk/callerid.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <asterisk/options.h>


static char *tdesc =
" PrepaidRate(): Returns the Tariff rate"
" for the destination number and credit of the caller.\n";
static char *app = "PrepaidRate";
static char *config = "prepaid.conf";
static char *synopsis = 
"Prepaid Call Rating Application";


static char *pghostname = NULL, *pgdbname = NULL, *pgdbuser = NULL, *pgpassword = NULL, *pgdbsock = NULL, *pgdbport = NULL;
static char *psqlauthorization = NULL;
static int hostname_alloc = 0, dbname_alloc = 0, dbuser_alloc = 0, password_alloc = 0, dbsock_alloc = 0, dbport_alloc = 0;
static int sqlauthorization_alloc = 0;
static int connected = 0; /* Initially not connected to the DB */

PGconn *conn;
ast_mutex_t postgreslock;

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

struct calldata {
    char cardnumber[256];
    char destination[256];
    char *credit;
    char *tariff;
    char *active;
    char *rate;
    char *tech;
    char *prefix;
    char *hostname;
    int timeout;
    char newdestination[256];
};

/* Checks if we are connected to the db - if not then it tries to reconnect */
static int db_check_connected(void)
{
	int i;
	PGresult *result;

	if (connected) {
                /* seems that we are connected - but let us check that */
                for (i=0; i<3; i++) {
                        /* PQstatus is not working as expected - so we are trying a query before PQstatus - hope that helps */
                        result = PQexec(conn, "SELECT 1");
                        PQclear(result);
                        if (PQstatus(conn) == CONNECTION_BAD) {
                                /* seems that we are having troubles with the connection - trying to reset the connection */
                                /* and set connected to false */
                                connected = 0;
                                ast_log(LOG_ERROR, "PrepaidRate: connection to database server '%s' is broken - I will try to reconnect (up to) 3 times\n", pghostname);
                                PQreset(conn);
                                /* retry it up to 3 times */
                                continue;
                        } else {
                                /* Everything is fine - so go on */
                                connected = 1;
                                break;
                        }
                }
        } else {
		/* We are not connected - so make the initial connection */
                conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
                if (PQstatus(conn) != CONNECTION_BAD) {
                        ast_log(LOG_DEBUG, "PrepaidRate: Successfully connected to database '%s'.\n", pgdbname);
                        connected = 1;
                } else {
                        ast_log(LOG_ERROR, "PrepaidRate: FATAL - Couldn't connect to database server '%s'. \n", pghostname);
                        connected = 0;
                }
        }
        return connected;
}

static int prepaidrate_dbrate(struct ast_channel *chan, struct calldata *cd)
{
	char sqlcmd[sqlcommandsize] = "";
	PGresult *result;


	if (db_check_connected()) {
		/* lookup rate */
	        sprintf(sqlcmd,psqlauthorization,cd->tariff,cd->destination);
        	ast_log(LOG_DEBUG,"PrepaidRate: SQL command as follows:  %s\n",sqlcmd);
	        ast_mutex_lock(&postgreslock);
        	result = PQexec(conn, sqlcmd);

		/* Check for Errors */
		if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
			ast_log(LOG_DEBUG,"PrepaidRate: FATAL ERROR: %s\n", PQresultErrorMessage(result));
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return -1;
	        }

                /* Check if there is a route for destination.
                   If not, add 100 to priority and return.
		   (Also there can be a SQL query problem)                   
                */

        	if(PQntuples(result)<=0) {
			ast_log(LOG_DEBUG,"PrepaidRate: WARNING: No destination route for \"%s\"\n", cd->destination);
			chan->priority += 100; 
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return -1;
	        }

                /* Check if we have the rate for destination.
                   If we don't , then add 200 to priority
                */

        	cd->rate = PQgetvalue(result,0,0);
	        if ( !cd->rate || !strlen(cd->rate) ) {
			ast_log(LOG_DEBUG,"PrepaidRate: WARNING: Destination \"%s\" is blocked.\n", cd->destination);
                        chan->priority += 200; 
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return -1;
	        }
		
		/* Check if rate is negative or not .
		  if it is, then add 300 to priority of the channel. 
		*/
        	if( atoi(cd->rate) <= 0 ) {
			ast_log(LOG_DEBUG,"PrepaidRate: WARNING: Negative rate for destination \"%s\"\n", cd->destination);
			chan->priority += 300;
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return -1;
	        }

		/* Get db values and assign them to call-detail struct*/		
 
        	cd->tech = PQgetvalue(result,0,1);
	        cd->prefix = PQgetvalue(result,0,2);
        	cd->hostname = PQgetvalue(result,0,3);
	        strncpy(cd->newdestination, PQgetvalue(result,0,4), sizeof(cd->newdestination));
	        ast_log(LOG_DEBUG,"PrepaidRate: SQL results: destination='%s', rate='%s', tech='%s', prefix='%s', hostname='%s'\n",cd->newdestination, cd->rate, cd->tech, cd->prefix, cd->hostname);
		PQclear(result);
	        ast_mutex_unlock(&postgreslock);

		return 0;

	} else {
		/* We are not connected to the DB */
		return -1;
	}
}


static int prepaidrate_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	struct localuser *u;
	struct calldata cd;
	char dialstr[256] = "";
	char destvar[20];
	char ratevar[20];
	int  timeout;
	char *credit;
	char timeoutvar[20];
	char ab_timeout[20];
        char argdata[256] = "";
        char *mydestination;
        char *mytariffplan;
        char *stringp;
        char *maxdigitstr;
        char *options;
	char tmp[256] = "";

	if (!data || ast_strlen_zero((char *)data)) {
		ast_log(LOG_WARNING, "PrepaidRate requires the arguments (Destination,Tariffplan)\n");
		return -1;
	}
	LOCAL_USER_ADD(u);

        credit=(char*) malloc(sizeof(char)*10);
	credit=pbx_builtin_getvar_helper(chan,"USERCREDIT");	
	if (!credit){
	  ast_log(LOG_WARNING,"Sorry You need USERCREDIT variable in your channel\n");
	}else{
	  
	}
        strncpy(argdata, (char *)data, sizeof(argdata) - 1);
        stringp=argdata;
        mydestination = strsep(&stringp, "|");
        mytariffplan = strsep(&stringp, "|");
        maxdigitstr = strsep(&stringp,"|");
        options = strsep(&stringp, "|");
        if (!(mytariffplan) || ast_strlen_zero(mytariffplan)) {
                ast_log(LOG_WARNING, "PrepaidRate requires an Tariffplan\n");
                return -1;
        }
        if (!(mydestination) || ast_strlen_zero(mydestination)) {
                ast_log(LOG_WARNING, "PrepaidRate requires an Destination\n");
                return -1;
        }
	strcpy(cd.destination, mydestination);
	cd.tariff = tmp;
	strcpy(cd.tariff, mytariffplan);

	res = prepaidrate_dbrate(chan, &cd);
	if (!res) {
                ast_cdr_reset(chan->cdr,0);

		snprintf(ratevar, sizeof(ratevar), "RATE");
		pbx_builtin_setvar_helper(chan, ratevar, cd.rate);
                ast_log(LOG_DEBUG,"PrepaidRate: '%s=%s'\n", ratevar, cd.rate);

                snprintf(dialstr, sizeof(dialstr), "%s/%s%s@%s", cd.tech, cd.prefix, cd.newdestination, cd.hostname);
		snprintf(destvar, sizeof(destvar), "DESTINATION1");
		pbx_builtin_setvar_helper(chan, destvar, dialstr);
     	        timeout=round(atoi(credit)/atoi(cd.rate));
     	        sprintf(ab_timeout,"%i",timeout);
     	        sprintf(timeoutvar,"%s", "TIMEOUT");
     	        pbx_builtin_setvar_helper(chan,timeoutvar,ab_timeout); 
     	        ast_log(LOG_WARNING,"PrepaidRate: %s=%s \n",timeoutvar,ab_timeout);
                ast_log(LOG_DEBUG,"PrepaidRate: '%s=%s'\n", destvar, dialstr);
	} else {
                
		ast_log(LOG_DEBUG, "PrepaidRate: No rate so exiting now... to priority '%i'\n",chan->priority);
		return 0;
	}

	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
        PQfinish(conn);
        connected = 0;
        if (pghostname && hostname_alloc) {
                free(pghostname);
                pghostname = NULL;
                hostname_alloc = 0;
        }
        if (pgdbname && dbname_alloc) {
                free(pgdbname);
                pgdbname = NULL;
                dbname_alloc = 0;
        }
        if (pgdbuser && dbuser_alloc) {
                free(pgdbuser);
                pgdbuser = NULL;
                dbuser_alloc = 0;
        }
        if (pgdbsock && dbsock_alloc) {
                free(pgdbsock);
                pgdbsock = NULL;
                dbsock_alloc = 0;
        }
        if (pgpassword && password_alloc) {
                free(pgpassword);
                pgpassword = NULL;
                password_alloc = 0;
        }
        if (pgdbport && dbport_alloc) {
                free(pgdbport);
                pgdbport = NULL;
                dbport_alloc = 0;
        }
        if (psqlauthorization && sqlauthorization_alloc) {
                free(psqlauthorization);
                psqlauthorization = NULL;
                sqlauthorization_alloc = 0;
        }

	return ast_unregister_application(app);
}

int load_module(void)
{

        int res;
        struct ast_config *cfg;
        struct ast_variable *var;
        char *tmp;

        cfg = ast_config_load(config);
        if (!cfg) {
                ast_log(LOG_WARNING, "Unable to load config for Prepaid Application: %s\n", config);
                return 0;
        }

        var = ast_variable_browse(cfg, "global");
        if (!var) {
                /* nothing configured */
                return 0;
        }

        tmp = ast_variable_retrieve(cfg,"global","hostname");
        if (tmp) {
                pghostname = malloc(strlen(tmp) + 1);
                if (pghostname != NULL) {
                        hostname_alloc = 1;
                        strcpy(pghostname, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database server hostname not specified.  Assuming localhost\n");
                pghostname = "localhost";
        }

        tmp = ast_variable_retrieve(cfg, "global", "dbname");
        if (tmp) {
                pgdbname = malloc(strlen(tmp) + 1);
                if (pgdbname != NULL) {
                        dbname_alloc = 1;
                        strcpy(pgdbname, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database not specified.  Assuming asterisk\n");
                pgdbname = "asterisk";
        }

        tmp = ast_variable_retrieve(cfg, "global", "user");
        if (tmp) {
                pgdbuser = malloc(strlen(tmp) + 1);
                if (pgdbuser != NULL) {
                        dbuser_alloc = 1;
                        strcpy(pgdbuser, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database user not specified.  Assuming postgres\n");
                pgdbuser = "postgres";
        }

        tmp = ast_variable_retrieve(cfg, "global", "password");
        if (tmp) {
                pgpassword = malloc(strlen(tmp) + 1);
                if (pgpassword != NULL) {
                        password_alloc = 1;
                        strcpy(pgpassword, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database password not specified.  Assuming blank\n");
                pgpassword = "";
        }

        tmp = ast_variable_retrieve(cfg,"global","port");
        if (tmp) {
                pgdbport = malloc(strlen(tmp) + 1);
                if (pgdbport != NULL) {
                        dbport_alloc = 1;
                        strcpy(pgdbport, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database port not specified.  Using 5432.\n");
                pgdbport = "5432";
        }

        tmp = ast_variable_retrieve(cfg, "sql", "authorization");
        if (tmp) {
                psqlauthorization = malloc(strlen(tmp) + 1);
                if (psqlauthorization != NULL) {
                        sqlauthorization_alloc = 1;
                        strcpy(psqlauthorization, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING, "Prepaid authorization query not specified. Using default.\n");
                psqlauthorization = "SELECT rate from countryprefix LEFT JOIN tariffrate USING (countrycode, subcode) WHERE tariffcode='%s' AND prefix=SUBSTRING('%s', 1, length(prefix)) ORDER BY LENGTH(prefix) desc";
        }

        ast_config_destroy(cfg);

        ast_log(LOG_DEBUG,"PrepaidRate: got hostname of %s\n", pghostname);
        ast_log(LOG_DEBUG,"PrepaidRate: got port of %s\n", pgdbport);
        if (pgdbsock) {
                ast_log(LOG_DEBUG,"PrepaidRate: got sock file of %s\n", pgdbsock);
	}
        ast_log(LOG_DEBUG,"PrepaidRate: got user of %s\n", pgdbuser);
        ast_log(LOG_DEBUG,"PrepaidRate: got dbname of %s\n", pgdbname);
        ast_log(LOG_DEBUG,"PrepaidRate: got password of %s\n", pgpassword);

        ast_log(LOG_DEBUG,"PrepaidRate: got sqlauthorization of '%s'\n", psqlauthorization);

        /* Check DB connection (it will automatically set up the connection) */
        db_check_connected();
	/* Initialise a mutex to make sure Postgres doesn't walk all over itself */
        ast_mutex_init(&postgreslock);

        res = ast_register_application(app, prepaidrate_exec, synopsis, tdesc);
        if (res) {
                ast_log(LOG_ERROR, "Unable to register PrepaidRate Application\n");
        }

        return res;

}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
