/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Prepaid Rate application
 * 
 * Copyright (C) 2004, Peter Nixon
 *
 * Peter Nixon <codemonkey@peternixon.net>
 *
 * $Id: app_prepaid_rateplan.c,v 1.4 2004/09/27 20:16:11 peternixon Exp $
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#define sqlcommandsize 2048

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <asterisk/utils.h>

#include <asterisk/app.h>
#include <asterisk/say.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>

#include <asterisk/config.h>
#include <pgsql/libpq-fe.h>

#include <asterisk/translate.h>
#include <asterisk/musiconhold.h>
#include <asterisk/callerid.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <asterisk/options.h>


static char *tdesc =
" PrepaidRatePlan(User): Returns the Tariff plan for the specified User."
" If no rateplan exists, it increments the Priority by 100\n";
static char *app = "PrepaidRatePlan";
static char *config = "prepaid.conf";
static char *synopsis = 
"Prepaid User Rating Application";


static char *pghostname = NULL, *pgdbname = NULL, *pgdbuser = NULL, *pgpassword = NULL, *pgdbsock = NULL, *pgdbport = NULL;
static char *psqluserauthorization = NULL , *pgusercredit = NULL;
static int hostname_alloc = 0, dbname_alloc = 0, dbuser_alloc = 0, password_alloc = 0, dbsock_alloc = 0, dbport_alloc = 0;
static int sqluserauthorization_alloc = 0 , usercredit_alloc = 0;
static int connected = 0; /* Initially not connected to the DB */

PGconn *conn;
ast_mutex_t postgreslock;

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

struct calldata {
    char cardnumber[256];
    char destination[256];
    char *credit;
    char *tariff;
    char *active;
    char *rate;
    char *tech;
    char *prefix;
    char *hostname;
    int timeout;
    char username[256];
    char tariffplan[256];
};

/* Checks if we are connected to the db - if not then it tries to reconnect */
static int db_check_connected(void)
{
	int i;
	PGresult *result;

	if (connected) {
                /* seems that we are connected - but let us check that */
                for (i=0; i<3; i++) {
                        /* PQstatus is not working as expected - so we are trying a query before PQstatus - hope that helps */
                        result = PQexec(conn, "SELECT 1");
                        PQclear(result);
                        if (PQstatus(conn) == CONNECTION_BAD) {
                                /* seems that we are having troubles with the connection - trying to reset the connection */
                                /* and set connected to false */
                                connected = 0;
                                ast_log(LOG_ERROR, "PrepaidRatePlan: connection to database server '%s' is broken - I will try to reconnect (up to) 3 times\n", pghostname);
                                PQreset(conn);
                                /* retry it up to 3 times */
                                continue;
                        } else {
                                /* Everything is fine - so go on */



                                connected = 1;
                                break;
                        }
                }
        } else {
		/* We are not connected - so make the initial connection */
                conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
                if (PQstatus(conn) != CONNECTION_BAD) {
                        ast_log(LOG_DEBUG, "PrepaidRatePlan: Successfully connected to database '%s'.\n", pgdbname);
                        connected = 1;
                } else {
                        ast_log(LOG_ERROR, "PrepaidRatePlan: FATAL - Couldn't connect to database server '%s'. \n", pghostname);
                        connected = 0;
                }
        }
        return connected;
}

static int prepaidrate_dbrate(struct ast_channel *chan, struct calldata *cd)
{
	char sqlcmd[sqlcommandsize] = "";
	PGresult *result;

	if (db_check_connected()) {
		/* lookup rate */
	        sprintf(sqlcmd,psqluserauthorization,cd->username);
        	ast_log(LOG_DEBUG,"PrepaidRatePlan: SQL command as follows:  %s\n",sqlcmd);
	        ast_mutex_lock(&postgreslock);
        	result = PQexec(conn, sqlcmd);

		/* Check for Errors */
		if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
			ast_log(LOG_ERROR,"PrepaidRatePlan: There was a problem with the SQL query: %s\n", PQresultErrorMessage(result));
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return -1;
	        }
        	if(PQntuples(result)<=0) {
                        chan->priority+=200;
			ast_log(LOG_WARNING,"PrepaidRatePlan: No Username '%s'. Jumping to Priority '%i'\n", cd->username, chan->priority);
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return 0;
	        }else{
    	           strncpy(cd->tariffplan, PQgetvalue(result,0,0), strlen( PQgetvalue(result,0,0) ));
		   cd->tariffplan[strlen(PQgetvalue(result,0,0))]  = '\0';
                }

	        if ( !cd->tariffplan || !strlen(cd->tariffplan) ) {
                        chan->priority+=300;
			ast_log(LOG_WARNING,"PrepaidRatePlan: No Tariff Plan for '%s'. Jumping to Priority '%i'\n", cd->username, chan->priority);
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
                        return 0;
	        }


                PQclear(result);

		/* lookup credit */
	        sprintf(sqlcmd,pgusercredit,cd->username);
        	ast_log(LOG_DEBUG,"PrepaidRatePlan: SQL command CREDIT as follows:  %s\n",sqlcmd);
        	result = PQexec(conn, sqlcmd);


		/* Check for Errors */

		if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
			ast_log(LOG_ERROR,"PrepaidRatePlan: There was a problem with the SQL query: %s\n", PQresultErrorMessage(result));
			PQclear(result);
			ast_mutex_unlock(&postgreslock);
			return -1;
	        }
        	if(PQntuples(result)<=0) {
			PQclear(result);
			chan->priority+=200;
			ast_log(LOG_WARNING,"PrepaidRatePlan: No Such User %s  So jumping to priority %i \n", cd->username,chan->priority);
			ast_mutex_unlock(&postgreslock);
                        return 0;
	        }else{
	        cd->credit=(char*)malloc(sizeof(char)*strlen(PQgetvalue(result,0,0)));
	        strncpy(cd->credit, PQgetvalue(result,0,0),strlen(PQgetvalue(result,0,0)));
		cd->credit[strlen(PQgetvalue(result,0,0))]  = '\0';
		}
		 
	        if ( !cd->credit || !strlen(cd->credit) || atoi(PQgetvalue(result,0,0)) <= 0 ) {
			//PQclear(result);
			//ast_mutex_unlock(&postgreslock);
                        if (ast_exists_extension(chan, chan->context, chan->exten, chan->priority + 100, chan->cid.cid_num)){
                            ast_log(LOG_WARNING,"context %s , extension %s ,priority %i ,callerid %s",chan->context, chan->exten , chan->priority +100, chan->cid.cid_num);
                        }
			chan->priority+=100;
			ast_log(LOG_WARNING,"PrepaidRatePlan: No or Low credit for User '%s' So jumping to priority %i \n", cd->username,chan->priority);
                         
	        }
	        
		PQclear(result);
	        ast_mutex_unlock(&postgreslock);
		return 0;

	} else {
		/* We are not connected to the DB */
		return -1;
	}
}


static int prepaidrateplan_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	struct localuser *u;
	struct calldata cd;
	char ratevar[20];
	char ucredit[20];

	LOCAL_USER_ADD(u);

        ast_log(LOG_WARNING,"callerid is -> %s\n",chan->cid.cid_num);
	strncpy(cd.username, chan->accountcode, strlen(chan->accountcode));
	cd.username[strlen(chan->accountcode)]  = '\0';
	res = prepaidrate_dbrate(chan, &cd);
	if (!res) {
                ast_cdr_reset(chan->cdr,0);

		snprintf(ratevar,sizeof(ratevar), "TARIFFPLAN");
		snprintf(ucredit,sizeof(ucredit), "USERCREDIT");
		
		pbx_builtin_setvar_helper(chan, ratevar, cd.tariffplan);
		pbx_builtin_setvar_helper(chan, ucredit, cd.credit);
                ast_log(LOG_WARNING,"PrepaidRatePlan: '%s=%s'\n", ratevar, cd.tariffplan);
                ast_log(LOG_WARNING,"PrepaidRatePlan: '%s=%s'\n", ucredit, cd.credit);
		ast_log(LOG_DEBUG, "PrepaidRatePlan: Here is the priority '%i' and res %i \n",chan->priority,res);
	} else {
		ast_log(LOG_DEBUG, "PrepaidRatePlan: No Tariff Plan. Exiting now...\n");
		return -1;
	}

	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
        PQfinish(conn);
        connected = 0;
        if (pghostname && hostname_alloc) {
                free(pghostname);
                pghostname = NULL;
                hostname_alloc = 0;
        }
        if (pgdbname && dbname_alloc) {
                free(pgdbname);
                pgdbname = NULL;
                dbname_alloc = 0;
        }
        if (pgdbuser && dbuser_alloc) {
                free(pgdbuser);
                pgdbuser = NULL;
                dbuser_alloc = 0;
        }
        if (pgdbsock && dbsock_alloc) {
                free(pgdbsock);
                pgdbsock = NULL;
                dbsock_alloc = 0;
        }
        if (pgpassword && password_alloc) {
                free(pgpassword);
                pgpassword = NULL;
                password_alloc = 0;
        }
        if (pgdbport && dbport_alloc) {
                free(pgdbport);
                pgdbport = NULL;
                dbport_alloc = 0;
        }
        if (pgusercredit && usercredit_alloc) {
                free(pgusercredit);
                pgusercredit = NULL;
                usercredit_alloc = 0;
        }
        if (psqluserauthorization && sqluserauthorization_alloc) {
                free(psqluserauthorization);
                psqluserauthorization = NULL;
                sqluserauthorization_alloc = 0;
        }

	return ast_unregister_application(app);
}

int load_module(void)
{

        int res;
        struct ast_config *cfg;
        struct ast_variable *var;
        char *tmp;

        cfg = ast_config_load(config);
        if (!cfg) {
                ast_log(LOG_WARNING, "Unable to load config for Prepaid Application: %s\n", config);
                return 0;
        }

        var = ast_variable_browse(cfg, "global");
        if (!var) {
                /* nothing configured */
                return 0;
        }

        tmp = ast_variable_retrieve(cfg,"global","hostname");
        if (tmp) {
                pghostname = malloc(strlen(tmp) + 1);
                if (pghostname != NULL) {
                        hostname_alloc = 1;
                        strcpy(pghostname, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database server hostname not specified.  Assuming localhost\n");
                pghostname = "localhost";
        }

        tmp = ast_variable_retrieve(cfg, "global", "dbname");
        if (tmp) {
                pgdbname = malloc(strlen(tmp) + 1);
                if (pgdbname != NULL) {
                        dbname_alloc = 1;
                        strcpy(pgdbname, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database not specified.  Assuming asterisk\n");
                pgdbname = "asterisk";
        }

        tmp = ast_variable_retrieve(cfg, "global", "user");
        if (tmp) {
                pgdbuser = malloc(strlen(tmp) + 1);
                if (pgdbuser != NULL) {
                        dbuser_alloc = 1;
                        strcpy(pgdbuser, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database user not specified.  Assuming postgres\n");
                pgdbuser = "postgres";
        }

        tmp = ast_variable_retrieve(cfg, "global", "password");
        if (tmp) {
                pgpassword = malloc(strlen(tmp) + 1);
                if (pgpassword != NULL) {
                        password_alloc = 1;
                        strcpy(pgpassword, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database password not specified.  Assuming blank\n");
                pgpassword = "";
        }

        tmp = ast_variable_retrieve(cfg,"global","port");
        if (tmp) {
                pgdbport = malloc(strlen(tmp) + 1);
                if (pgdbport != NULL) {
                        dbport_alloc = 1;
                        strcpy(pgdbport, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"Prepaid database port not specified.  Using 5432.\n");
                pgdbport = "5432";
        }
        
        tmp = ast_variable_retrieve(cfg,"sql","user_credit");
        if (tmp) {
                pgusercredit = malloc(strlen(tmp) + 1);
                if (pgusercredit != NULL) {
                        usercredit_alloc = 1;
                        strcpy(pgusercredit, tmp);
                } else {
                        ast_log(LOG_ERROR,"Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING,"User Credit query not specified.  Using default.\n");
                pgusercredit = "SELECT credit FROM friends WHERE name='%s';";
        }

        tmp = ast_variable_retrieve(cfg, "sql", "userauthorization");
        if (tmp) {
                psqluserauthorization = malloc(strlen(tmp) + 1);
                if (psqluserauthorization != NULL) {
                        sqluserauthorization_alloc = 1;
                        strcpy(psqluserauthorization, tmp);
                } else {
                        ast_log(LOG_ERROR, "Out of memory error.\n");
                        return -1;
                }
        } else {
                ast_log(LOG_WARNING, "Prepaid authorization query not specified. Using default.\n");
                psqluserauthorization = "SELECT * FROM asterisk_authorize_user('%s') AS authorize_user(tariffplan text, activated bool);";
        }

        ast_config_destroy(cfg);

        ast_log(LOG_DEBUG,"PrepaidRate: got hostname of %s\n", pghostname);
        ast_log(LOG_DEBUG,"PrepaidRate: got port of %s\n", pgdbport);
        if (pgdbsock) {
                ast_log(LOG_DEBUG,"PrepaidRate: got sock file of %s\n", pgdbsock);
	}
        ast_log(LOG_DEBUG,"PrepaidRate: got user of %s\n", pgdbuser);
        ast_log(LOG_DEBUG,"PrepaidRate: got dbname of %s\n", pgdbname);
        ast_log(LOG_DEBUG,"PrepaidRate: got password of %s\n", pgpassword);

        ast_log(LOG_DEBUG,"PrepaidRate: got sqlauthorization of '%s'\n", psqluserauthorization);

        /* Check DB connection (it will automatically set up the connection) */
        db_check_connected();
	/* Initialise a mutex to make sure Postgres doesn't walk all over itself */
        ast_mutex_init(&postgreslock);

        res = ast_register_application(app, prepaidrateplan_exec, synopsis, tdesc);
        if (res) {
                ast_log(LOG_ERROR, "Unable to register PrepaidRate Application\n");
        }

        return res;

}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
